@@ -1,5 +1,5 @@
 ;;  -*-  indent-tabs-mode:nil; coding: utf-8 -*-
-;;  Copyright (C) 2013,2014,2015,2016
+;;  Copyright (C) 2013,2014,2015,2016,2017,2018,2019
 ;;      "Mu Lei" known as "NalaGinrut" <NalaGinrut@gmail.com>
 ;;  Artanis is free software: you can redistribute it and/or modify
 ;;  it under the terms of the GNU General Public License and GNU
@@ -59,6 +59,14 @@
 (define (->n n m)
   (format #f "~a(~{'~a'~^,~}) ~{~a~^ ~}" n (car m) (cdr m)))
 
+(define (float-args-is-valid? args)
+  (if (< (length args) 2)
+      #f
+      (let ((i (car args))
+            (f (cadr args)))
+        (and (positive-integer? i)
+             (positive-integer? f)))))
+
 (define (->mysql-type name . args)
   (case name
     ;; Text types:
@@ -101,11 +109,19 @@
     ;; FLOAT(size,d) A small number with a floating decimal point.
     ;; The maximum number of digits may be specified in the size parameter.
     ;; The maximum number of digits to the right of the decimal point is specified in the d parameter
-    ((float) (->2 name args))
+    ((float)
+     (if (float-args-is-valid? args)
+         (->2 name args)
+         (throw 'artanis-err 500 ->mysql-type
+                "Invalid definition: `~a'" (->2 name args))))
     ;; DOUBLE(size,d) A large number with a floating decimal point.
     ;; The maximum number of digits may be specified in the size parameter.
     ;; The maximum number of digits to the right of the decimal point is specified in the d parameter
-    ((double) (->2 name args))
+    ((double)
+     (if (float-args-is-valid? args)
+         (->2 name args)
+         (throw 'artanis-err 500 ->mysql-type
+                "Invalid definition: `~a'" (->2 name args))))
     ;; DECIMAL(size,d) A DOUBLE stored as a string , allowing for a fixed decimal point.
     ;; The maximum number of digits may be specified in the size parameter.
     ;; The maximum number of digits to the right of the decimal point is specified in the d parameter
@@ -134,7 +150,7 @@
     ;; NOTE: Values allowed in four-digit format: 1901 to 2155.
     ;;       Values allowed in two-digit format: 70 to 69, representing years from 1970 to 2069
     ((year) (->0 name args))
-    (else (throw 'artanis-err 500 "->mysql-type: Invalid type name" name))))
+    (else (throw 'artanis-err 500 ->mysql-type "Invalid type name `~a'" name))))
 
 (define (->postgresql-type name . args)
   (case name
@@ -146,7 +162,7 @@
     ((int) (->0 name args))
 
     ;; Character Types
-    
+
     ;; char for 1 byte single-byte internal type
     ((char) (if (null? args) (->0 name args) (->1 name args)))
     ((name) (->0 name args)) ; 64 bytes internal type for object names
@@ -207,7 +223,7 @@
     ;; Composite Types
     ;; TODO: do we really need this?
 
-    (else (throw 'artanis-err 500 "->postgresql-type: Invalid type name" name))))
+    (else (throw 'artanis-err 500 ->postgresql-type "Invalid type name `~a'" name))))
 
 (define (->sql-general-type name . args)
   (case name
@@ -231,7 +247,11 @@
     ((numeric) (->2 name args)) ; NUMERIC(p,s) Exact numerical, precision p, scale s. (Same as DECIMAL)
     ;; FLOAT(p) Approximate numerical, mantissa precision p. A floating number in base 10 exponential notation.
     ;; The size argument for this type consists of a single number specifying the minimum precision
-    ((float) (if (null? args) (->0 name args) (->1 name args)))
+    ((float)
+     (cond
+      ((null? args) (->0 name args))
+      ((float-args-is-valid? args) #f) ; trigger DBD specific float definition
+      (else (->1 name args))))
     ((real) (->0 name args)) ; REAL Approximate numerical, mantissa precision 7
     ;; DOUBLE PRECISION Approximate numerical, mantissa precision 16
     ((double-precision) (->0 "double precision" args))
@@ -251,7 +271,7 @@
   (cond
    ((string? x) (string->symbol x))
    ((symbol? x) x)
-   (else (throw 'artanis-err 500 "->symbol: Invalid type!" x))))
+   (else (throw 'artanis-err 500 '->symbol "Invalid type!" x))))
 
 ;; (map (lambda (x) (->sql-type (cdr x)))
 ;;      '((name varchar 10) (age int 5) (email varchar 255)))
@@ -259,14 +279,14 @@
 (define-macro (->sql-type name-and-args)
   `(or (apply ->sql-general-type ,name-and-args)
        (apply ,(symbol-append '-> (->symbol (get-conf '(db dbd))) '-type)
-        ,name-and-args)))
+              ,name-and-args)))
 
 (define (make-table-dropper rc/conn)
   (define conn
-    (cond 
-     ((route-context? rc/conn) (rc-conn rc/conn))
+    (cond
+     ((route-context? rc/conn) (DB-open rc/conn))
      ((<connection>? rc/conn) rc/conn)
-     (else (throw 'artanis-err 500 "make-table-dropper: Invalid rc or conn!" rc/conn))))
+     (else (throw 'artanis-err 500 make-table-dropper "Invalid rc or conn!" rc/conn))))
   (lambda* (name #:key (dump #f))
     (let ((sql (->sql drop table if exists name)))
       (cond
@@ -274,37 +294,45 @@
        (else sql)))))
 
 (define *exception-opts*
-  '(#:no-edit))
+  '(#:no-edit #:default #:comment #:storage))
 (define (is-exception-opt? x)
   (memq x *exception-opts*))
 
 (define (->mysql-opts x opts)
-  (define-syntax-rule (->value x)
-    (format #f "~:@(~a~) ~a" x (kw-arg-ref opts x)))
+  (define-syntax-rule (->value o x)
+    (format #f "~:@(~a~) ~s" o (kw-arg-ref opts x)))
   (case x
     ((#:not-null) "NOT NULL")
     ((#:null) "NULL")
-    ((#:default) (->value x))
+    ((#:default) (->value 'default x))
     ((#:unique) "UNIQUE")
     ((#:unique-key) "UNIQUE KEY")
     ((#:primary-key) "PRIMARY KEY")
     ((#:key) "KEY")
     ((#:auto-increment) "AUTO_INCREMENT")
-    ((#:comment) (->value x))
+    ((#:comment) (->value 'comment x))
     ((#:column-format) "COLUMN_FORMAT")
-    ((#:storage) (->value x))
+    ((#:storage) (->value 'storage x))
     ((#:reference-definition) "reference_definition")
     (else
-     (if (is-exception-opt? x)
-         ""
-         (throw 'artanis-err 500 "Invalid opts for MySQL table definition!" x)))))
+     (cond
+      ((keyword? x)
+       (if (is-exception-opt? x)
+           ""
+           ;; Throw exception for invalid keyword
+           (throw 'artanis-err 500 ->mysql-opts
+                  "Invalid opts `~a' for MySQL table definition!" x)))
+      (else
+       ;; Just ignore the non-keyword, since there're only 2 situations:
+       ;; 1. The value of option has already been fetched.
+       ;; 2. Invalid item here, neither keyword as options, nor valid value.
+       "")))))
 
 (define (->postgresql-opts dbd opts)
-  (format #t "PostgreSQL migration hasn't been supported yet!~%")
-  #t)
+  (format #t "PostgreSQL migration hasn't been supported yet!~%"))
+
 (define (->sqlite3-opts dbd opts)
-  (format #t "SQLite3 migration hasn't been supported yet!~%")
-  #t)
+  (format #t "SQLite3 migration hasn't been supported yet!~%"))
 
 (define *table-builder-opts-handler*
   `((mysql . ,->mysql-opts)
@@ -323,9 +351,10 @@
 (define* (make-table-builder rc/conn)
   (define conn
     (cond
-     ((route-context? rc/conn) (rc-conn rc/conn))
+     ((route-context? rc/conn) (DB-open rc/conn))
      ((<connection>? rc/conn) rc/conn)
-     (else (throw 'artanis-err 500 "make-table-builder: Invalid rc or conn!" rc/conn))))
+     (else (throw 'artanis-err 500 make-table-builder
+                  "Invalid rc or conn `~a'!" rc/conn))))
   (define (table-drop! tname)
     (DB-query conn (->sql drop table if exists tname)))
   (define (->opts opts)
@@ -341,25 +370,31 @@
     (match x
       ((types ... (opts ...)) (values types (if (null? opts) "" (->opts opts))))
       ((types ...) (values types ""))
-      (else (throw 'artanis-err 500 "->type/opts: invalid definition of the table!" x))))
+      (else (throw 'artanis-err 500 ->type/opts
+                   "Invalid definition of the table `~a'!" x))))
   (define (->types x)
     (call-with-values
         (lambda () (->type/opts x))
       (lambda (types opts)
         (format #f "~{~a~^ ~}" (list (car types) (->sql-type (cdr types)) opts)))))
+  (define (gen-primary-keys primary-keys)
+    (if (null? primary-keys)
+        ""
+        (format #f "PRIMARY KEY (~{~a~^,~})" primary-keys)))
   ;; TODO: We need a mechanism to sync tables constrained by foreign-keys, since some DB doesn't
   ;;       support foreign keys directly, so we have to provide it outside.
   ;; TODO: who to deal with constrained tables without foreign-keys in stateless?
   (lambda* (tname defs #:key (if-exists? #f) (engine (get-conf '(db engine)))
-                  (dump #f) (primary-keys '()))                  
+                  (dump #f) (primary-keys '()))
     (let* ((types (map ->types defs))
+           (pks (gen-primary-keys primary-keys))
            (sql (case if-exists?
                   ((overwrite drop)
                    (table-drop! tname)
-                   (->sql create table tname (types) engine))
+                   (->sql create table tname (types pks) engine))
                   ((ignore)
-                   (->sql create table if not exists tname (types) engine))
-                  (else (->sql create table tname (types) engine)))))
+                   (->sql create table if not exists tname (types pks) engine))
+                  (else (->sql create table tname (types pks) engine)))))
       (cond
        ((not dump)
         (DB-query conn sql)
@@ -372,7 +407,7 @@
             ('(drop-primary-keys)
              (DB-query conn (->sql alter table tname drop primary key)))
             ;; TODO
-            (_ (throw 'artanis-err 500 "make-table-builder: Invalid cmd!" cmd)))))
+            (_ (throw 'artanis-err 500 make-table-builder "Invalid cmd `~a'!" cmd)))))
        (else sql)))))
 
 ;; make-table-setter could be mapped to UPDATE or INSERT, depends on condition.
@@ -394,21 +429,28 @@
         (((? string? wcond))
          (lp (cdr next) kvs wcond))
         (() (values kvs w))
-        (else (throw 'artanis-err 500 "->kvp: invalid kargs" next)))))
+        (else (throw 'artanis-err 500 make-table-setter "Invalid kargs `~a'" next)))))
   (define (->kv kvp) (srfi-1:unzip2 kvp))
+  (define (guard vl)
+    (map (lambda (s)
+           ;; TODO: drop unsafe substrings
+           s)
+         vl))
   (lambda (tname . kargs)
     (let-values (((kvp wcond) (->kvp kargs)))
       (let ((sql (if (string-null? wcond)
                      (let-values (((k v) (->kv kvp)))
-                       (->sql insert into tname k values v))
+                       (let ((sv (guard v)))
+                         (->sql insert into tname k values sv)))
                      ;; NOTE: If there's no cond string (say, `where' clauses), you MUSTN'T
                      ;;       use UPDATE because it'll effect all the records!!! In such case,
                      ;;       you should use INSERT.
                      (->sql update tname set kvp wcond)))
-            (conn (cond 
-                   ((route-context? rc/conn) (rc-conn rc/conn))
+            (conn (cond
+                   ((route-context? rc/conn) (DB-open rc/conn))
                    ((<connection>? rc/conn) rc/conn)
-                   (else (throw 'artanis-err 500 "make-table-setter: Invalid rc or conn!" rc/conn)))))
+                   (else (throw 'artanis-err 500 make-table-setter
+                                "Invalid rc or conn `~a'!" rc/conn)))))
         (DB-query conn sql)))))
 
 (define (make-table-getter rc/conn)
@@ -421,12 +463,12 @@
   (define (->group-by group-by)
     (match group-by
       ((? list columns)
-       (format #f " ~{~a~^,~} " columns))
+       (format #f " group by ~{~a~^,~} " columns))
       (_ #f)))
   (define (->order-by order-by)
     (match order-by
       ((columns ... (? (cut memq <> '(asc desc)) m))
-       (format #f " ~{~a~^,~} ~a " columns m))
+       (format #f " order by ~{~a~^,~} ~a " columns m))
       (_ #f)))
   (define (->opts ret group-by order-by cnd foreach)
     (define-syntax-rule (-> x tox)
@@ -434,9 +476,9 @@
     (define-syntax-rule (cond-combine c lst)
       (cond
        ((not (string? c))
-        (throw 'artanis-err 500 "Invalid #:condition" c))
+        (throw 'artanis-err 500 make-table-getter "Invalid #:condition `~a'" c))
        ((not (list? lst))
-        (throw 'artanis-err 500 "Invalid #:foreach" lst))
+        (throw 'artanis-err 500 make-table-getter "Invalid #:foreach `~a'" lst))
        (else
         (match lst
           (() c)
@@ -445,13 +487,13 @@
                    (if (string-null? c) " where " (string-append c " and "))
                    column vals))
           (_
-           (throw 'artanis-err 500
-                  "Invalid #:foreach, should be (column (val1 val2 val3 ...))"
+           (throw 'artanis-err 500 make-table-getter
+                  "Invalid #:foreach `~a', should be (column (val1 val2 val3 ...))"
                   lst))))))
     (string-concatenate
-     (list (-> group-by ->group-by)
+     (list (cond-combine cnd foreach)
+           (-> group-by ->group-by)
            (-> order-by ->order-by)
-           (cond-combine cnd foreach)
            (-> ret ->ret))))
   (define (->mix columns functions)
     `(,@columns ,@(map (lambda (f) (format #f "~a(~{~a~^,~})" (car f) (cdr f))) functions)))
@@ -488,19 +530,31 @@
                   ;; #:foreach '(city ("sz" "bj" "sh"))
                   ;; NOTE: This is useful to avoid N+1 query problem.
                   (dump #f)
-                  ;; Dump SQL to a string as result, when #:dump set to #t, it won't do query operation. 
+                  ;; Dump SQL to a string as result, when #:dump set to #t, it won't do query operation.
+                  (mode 'raw)
+                  ;; The return value;
+                  ;; 1. 'raw: returns all rows
+                  ;; 2. 'getter: return a getter as a function
                   )
     (let ((sql (format #f "select ~{~a~^,~} from ~a ~a;"
                        (->mix columns functions) tname (->opts ret group-by order-by condition foreach)))
           (conn (cond
-                 ((route-context? rc/conn) (rc-conn rc/conn))
+                 ((route-context? rc/conn) (DB-open rc/conn))
                  ((<connection>? rc/conn) rc/conn)
-                 (else (throw 'artanis-err 500
-                              "make-table-getter: Invalid rc or conn!" rc/conn)))))
+                 (else (throw 'artanis-err 500 make-table-getter
+                              "Invalid rc or conn `~a'!" rc/conn)))))
       (cond
        ((not dump)
         (DB-query conn sql)
-        (DB-get-all-rows conn))
+        (let ((ret (DB-get-all-rows conn)))
+          (case mode
+            ((raw) ret)
+            ((getter)
+             (lambda (k)
+               (and (not (null? ret))
+                    (assoc-ref (car ret) k))))
+            (else (throw 'artanis-err 500 make-table-getter
+                         "Invalid mode `~a'" mode)))))
        (else sql)))))
 
 (define (make-table-modifier rc/conn)
@@ -512,30 +566,39 @@
     (case (get-conf '(db dbd))
       ((mysql) (->sql alter table tname modify column col t))
       ((postgresql) (->sql alter table tname alter column col t))
-      ((sqlite3) (throw 'artanis-err 500 "SQLite3 doesn't support table modification!"))
-      (else (throw 'artanis-err 500 "table-alter: Unsupported DBD!" (get-conf '(db dbd))))))
+      ((sqlite3) (throw 'artanis-err 500 table-alter
+                        "SQLite3 doesn't support table modification!"))
+      (else (throw 'artanis-err 500 table-alter " Unsupported DBD `~a'!"
+                   (get-conf '(db dbd))))))
   (define (table-rename tname new)
     (case (get-conf '(db dbd))
       ((mysql) (->sql alter table tname rename new))
       ((postgresql sqlite3) (->sql alter table tname rename to new))
-      (else (throw 'artanis-err 500 "table-alter: Unsupported DBD!" (get-conf '(db dbd))))))
+      (else (throw 'artanis-err 500 table-rename "Unsupported DBD `~a'!"
+                   (get-conf '(db dbd))))))
   (define (column-drop tname cname)
     (case (get-conf '(db dbd))
       ((mysql postgresql) (->sql alter table tname drop column cname))
-      ((sqlite3) (throw 'artanis-err 500 "SQLite3 doesn't support table column modification!"))
-      (else (throw 'artanis-err 500 "table-alter: Unsupported DBD!" (get-conf '(db dbd))))))
+      ((sqlite3) (throw 'artanis-err 500 column-drop
+                        "SQLite3 doesn't support table column modification!"))
+      (else (throw 'artanis-err 500 column-drop
+                   "Unsupported DBD!" (get-conf '(db dbd))))))
   (define (column-rename tname old new . type)
     (case (get-conf '(db dbd))
       ((mysql) (let ((t (car type))) (->sql alter table tname change column old new t)))
       ((postgresql) (->sql alter table tname rename column old new))
-      ((sqlite3) (throw 'artanis-err 500 "SQLite3 doesn't support table column modification!"))
-      (else (throw 'artanis-err 500 "table-alter: Unsupported DBD!" (get-conf '(db dbd))))))
+      ((sqlite3) (throw 'artanis-err 500 column-rename
+                        "SQLite3 doesn't support table column modification!"))
+      (else (throw 'artanis-err 500 column-rename
+                   "Unsupported DBD `~a'!" (get-conf '(db dbd))))))
   (define (index-rename tname old new)
     (case (get-conf '(db dbd))
       ((mysql) (->sql alter table tname index old rename to new))
       ((postgresql) (->sql alter table tname index old rename to new))
-      ((sqlite3) (throw 'artanis-err 500 "SQLite3 doesn't support table column modification!"))
-      (else (throw 'artanis-err 500 "table-alter: Unsupported DBD!" (get-conf '(db dbd))))))
+      ((sqlite3) (throw 'artanis-err 500 index-rename
+                        "SQLite3 doesn't support table column modification!"))
+      (else (throw 'artanis-err 500 index-rename
+                   "Unsupported DBD `~a'!" (get-conf '(db dbd))))))
   (define (gen-sql tname op args)
     (case op
       ((add) (apply table-add tname args))
@@ -545,23 +608,25 @@
       ((rename) (apply table-rename tname args))
       ((column-rename) (apply column-rename tname args))
       ((index-rename) (apply index-rename args))
-      (else (throw 'artanis-err 500 "make-table-modifier: Invalid op!" op))))
+      (else (throw 'artanis-err 500 make-table-modifier
+                   "Invalid op `~a'!" op))))
   (lambda (tname op . args)
     (let ((sql (gen-sql tname op args))
-          (conn (cond 
-                 ((route-context? rc/conn) (rc-conn rc/conn))
+          (conn (cond
+                 ((route-context? rc/conn) (DB-open rc/conn))
                  ((<connection>? rc/conn) rc/conn)
-                 (else (throw 'artanis-err 500
-                              "make-table-modifier: Invalid rc or conn!" rc/conn)))))
+                 (else (throw 'artanis-err 500 make-table-modifier
+                              "Invalid rc or conn `~a'!" rc/conn)))))
       (DB-query conn sql))))
 
 ;; NOTE: the name of columns is charactar-caseless, at least in MySQL/MariaDB.
 (define (map-table-from-DB rc/conn)
   (define conn
     (cond
-     ((route-context? rc/conn) (rc-conn rc/conn))
+     ((route-context? rc/conn) (DB-open rc/conn))
      ((<connection>? rc/conn) rc/conn)
-     (else (throw 'artanis-err 500 "map-table-from-DB: invalid rc/conn" rc/conn))))
+     (else (throw 'artanis-err 500 map-table-from-DB
+                  "Invalid rc/conn `~a'" rc/conn))))
   (define getter (make-table-getter conn))
   (define setter (make-table-setter conn))
   (define builder (make-table-builder conn))
@@ -577,7 +642,7 @@
   ;; have chosen Scheme programming language for hacking.
   ;; PS: I'm not boasting that the whole Artanis would be stateless, but FPRM should do it as possible.
   ;; I maybe wrong and fail, but it's worth to try.
-  (define (get-table-schema tname)      
+  (define (get-table-schema tname)
     (let* ((sql (->sql select '("lcase(column_name)") from
                        (select * from 'information_schema.columns (having #:table_name tname))
                        as 'tmp_alias))
@@ -591,6 +656,7 @@
     (and (srfi-1:every (lambda (x) (memq x schema)) (-> args)) #t))
   (define (table-exists? tname)
     (DB-query rc/conn (format #f "show tables like '~a';" tname))
+    ;; If there's no result, dbi will return #f as the result.
     (DB-get-top-row rc/conn))
   (lambda (cmd tname . args)
     (define-syntax-rule (->call func)
@@ -613,14 +679,14 @@
       ;; schema is always in downcase.
       ((schema) (get-table-schema tname))
       ((mod) (->call modifier))
-      (else (throw 'artanis-err 500 "map-table-from-DB: Invalid cmd" cmd)))))
+      (else (throw 'artanis-err 500 map-table-from-DB
+                   "Invalid cmd: `~a'" cmd)))))
 
 (define* (fprm-find mt tname val #:key (id 'id))
   (mt 'get tname #:condition (where (symbol->keyword id) val)))
 
 (define *table-mapper-handlers*
-  `((find . ,fprm-find)
-    ))
+  `((find . ,fprm-find)))
 
 (define (create-table-mapper tname rc/conn)
   (let ((mt (map-table-from-DB rc/conn)))
