@@ -1,5 +1,5 @@
 ;;  -*-  indent-tabs-mode:nil; coding: utf-8 -*-
-;;  Copyright (C) 2013,2014,2015,2016,2017
+;;  Copyright (C) 2013,2014,2015,2016,2017,2018
 ;;      "Mu Lei" known as "NalaGinrut" <NalaGinrut@gmail.com>
 ;;  Artanis is free software: you can redistribute it and/or modify
 ;;  it under the terms of the GNU General Public License and GNU
@@ -22,12 +22,14 @@
   #:use-module (artanis env)
   #:use-module (artanis route)
   #:use-module (artanis db)
+  #:use-module (artanis lpc)
   #:use-module (artanis fprm)
   #:use-module (artanis config)
-  #:use-module (srfi srfi-9)
+  #:use-module (artanis third-party redis)
   #:use-module ((rnrs) #:select (define-record-type))
   #:use-module (web request)
   #:use-module (ice-9 format)
+  #:use-module (ice-9 match)
   #:export (session-set!
             session-ref
             session-expired?
@@ -76,8 +78,11 @@
     ret))
 
 (define (get-session-file sid)
-  (format #f "~a/prv/~a/~a.session" (current-toplevel)
-          (get-conf '(session path)) sid))
+  (let ((ct (current-toplevel)))
+    (if ct
+        (format #f "~a/prv/~a/~a.session" (current-toplevel)
+                (get-conf '(session path)) sid)
+        (format #f "session/~a.session" sid))))
 
 (define (new-session rc data expires)
   (let ((expires-str (make-expires expires))
@@ -86,9 +91,6 @@
                     ("client"  . ,ip)
                     ("data"    . ,data))))) ; data is assoc list
 
-;; TODO: Support session-engine:
-;; session.engine = redis or memcached, for taking advantage of k-v-DB.
-
 ;; TODO: session key-values should be flushed into set-cookie in rc, and should be encoded
 ;;       with base64.
 
@@ -97,12 +99,68 @@
    name ; symbol
    meta ; anything necessary for a specific backend
    init ; -> session-backend
-   store! ; -> session-backend -> string -> hash-table 
+   store! ; -> session-backend -> string -> hash-table
    destory! ; -> session-backend -> string
    restore ; -> session-backend -> string
    set! ; -> session-backend -> string -> string -> object
    ref)) ; -> session-backend -> string -> string
 
+;; Support session-engine:
+;; session.engine = redis or memcached, for taking advantage of k-v-DB.
+(define (backend:session-init/redis sb)
+  (DEBUG "Init session redis backend is done!  ~%" sb))
+
+(define (backend:session-store/redis sb sid ss)
+  (let ((redis (session-backend-meta sb))
+        (s (object->string (session->alist ss))))
+    (backend-impl:set!/redis redis sid s)))
+
+(define (backend:session-destory/redis sb sid)
+  (let ((redis (session-backend-meta sb)))
+    (if (backend:session-restore/redis sb sid)
+        (backend-impl:remove!/redis redis sid))
+    (backend-impl:destroy!/redis redis)))
+
+(define (backend:session-restore/redis sb sid)
+  (let* ((redis (session-backend-meta sb))
+         (line (backend-impl:ref/redis redis sid))
+         (ss (and line
+                  (and=> (call-with-input-string line read) make-session))))
+    ss))
+
+(define (backend:session-set/redis sb sid k v)
+  (let ((redis (session-backend-meta sb)))
+    (cond
+     ((backend:session-restore/redis sb sid)
+      => (lambda (ss)
+           (hash-set! ss k v)
+           (backend-impl:set!/redis
+            redis
+            sid
+            (object->string (session->alist ss)))))
+     (else
+      (throw 'artanis-err 500 backend:session-restore/redis
+             "Session id (~a) doesn't hit anything!~%" sid)))))
+
+(define (backend:session-ref/redis sb sid k)
+  (let ((redis (session-backend-meta sb)))
+    (cond
+     ((backend:session-restore/redis sb sid)
+      => (lambda (ss) (assoc-ref (hash-ref ss "data") k)))
+     (else
+      (throw 'artanis-err 500 backend:session-ref/redis
+             "Session id (~a) doesn't hit anything!~%" sid)))))
+
+(define* (new-session-backend/redis #:key (host "127.0.0.1") (port 6379))
+  (make-session-backend 'redis
+                        (new-lpc-backend/redis #:host host  #:port port)
+                        backend:session-init/redis
+                        backend:session-store/redis
+                        backend:session-destory/redis
+                        backend:session-restore/redis
+                        backend:session-set/redis
+                        backend:session-ref/redis))
+
 ;; session.engine = db, for managing sessions with DB support.
 (define (backend:session-init/db sb)
   (DEBUG "Initilizing session backend `~:@(~a~)'...~%" 'db)
@@ -189,7 +247,7 @@
     => (lambda (ss) (hash-set! ss k v)))
    (else
     (throw 'artanis-err 500 backend:session-restore/simple
-     (format #f "Session id (~a) doesn't hit anything!~%" sid)))))
+           "Session id (~a) doesn't hit anything!~%" sid))))
 
 (define (backend:session-ref/simple sb sid k)
   (cond
@@ -197,7 +255,7 @@
     => (lambda (ss) (hash-ref (hash-ref ss "data") k)))
    (else
     (throw 'artanis-err 500 backend:session-ref/simple
-           (format #f "Session id (~a) doesn't hit anything!~%" sid)))))
+           "Session id (~a) doesn't hit anything!~%" sid))))
 
 (define (new-session-backend/simple)
   (make-session-backend 'simple
@@ -233,8 +291,8 @@
       (DEBUG "Session path `~a' exists, keep it for existing sessions!~%" path))
      (else
       (throw 'artanis-err 500 backend:session-init/file
-             (format #f "Session path `~a' conflict with an existed file!~%"
-                     path))))))
+             "Session path `~a' conflict with an existed file!~%"
+             path)))))
 
 (define (backend:session-store/file sb sid ss)
   (let ((s (session->alist ss)))
@@ -290,7 +348,7 @@
   ((session-backend-set! (current-session-backend))
    (current-session-backend)
    k v))
-  
+
 (define (session-ref sid k)
   ((session-backend-ref (current-session-backend))
    (current-session-backend)
@@ -308,16 +366,17 @@
 (define (session-restore sid)
   (let ((session ((session-backend-restore (current-session-backend))
                   (current-session-backend) sid)))
-    (and session
-         (cond
-          ((session-expired? session)
-           (DEBUG "[Session] sid: ~a is expired, destory!~%" sid)
-           (session-destory! sid)
-           #f) ; expired then return #f
-          (else
-           (DEBUG "[Session] Restored session: ~a~%"
-                  (hash-map->list cons session))
-           session))))) ; non-expired, return session
+    (if session
+        (cond
+         ((session-expired? session)
+          (DEBUG "[Session] sid: ~a is expired, destory!~%" sid)
+          (session-destory! sid)
+          'expired) ; expired then return #f
+         (else
+          (DEBUG "[Session] Restored session: ~a~%"
+                 (hash-map->list cons session))
+          session)) ; non-expired, return session
+        'not-found)))
 ;; if no session then return #f
 
 (define (session-store! sid session)
@@ -327,8 +386,12 @@
   session)
 
 (define* (session-spawn rc #:key (data '()) (expires 3600))
+  (define (is-valid-session? s)
+    (case s
+      ((expired not-found) #f)
+      (else s)))
   (let* ((sid (get-new-sid))
-         (session (or (session-restore sid)
+         (session (or (is-valid-session? (session-restore sid))
                       (session-store! sid (new-session rc data expires)))))
     (DEBUG "Session spawned: sid - ~a, data - ~a~%" sid data)
     (values sid session)))
@@ -336,18 +399,26 @@
 (define *session-backend-table*
   `((simple . ,new-session-backend/simple)
     (db     . ,new-session-backend/db)
-    (file   . ,new-session-backend/file)))
+    (file   . ,new-session-backend/file)
+    (redis  . ,new-session-backend/redis)))
 
 (define (add-new-session-backend name maker)
   (set! *session-backend-table*
-        (assoc-set! *session-backend-table* name maker)))
+    (assoc-set! *session-backend-table* name maker)))
+
+(define (create-new-session conf)
+  (match conf
+    (('redis host port)
+     (lambda ()
+       ((assoc-ref *session-backend-table* 'redis) #:host host #:port port)))
+    (else (assoc-ref *session-backend-table* conf))))
 
 (define (session-init)
-  (cond
-   ((assoc-ref *session-backend-table* (get-conf '(session backend)))
-    => (lambda (maker)
-         (let ((sb (maker)))
-           ((session-backend-init sb) sb)
-           (change-session-backend! sb))))
-   (else (error (format #f "Invalid session backdend: ~:@(~a~)"
-                        (get-conf '(session backend)))))))
+  (let ((conf (get-conf '(session backend))))
+    (cond
+     ((create-new-session conf)
+      => (lambda (maker)
+           (let ((sb (maker)))
+             ((session-backend-init sb) sb)
+             (change-session-backend! sb))))
+     (else (error (format #f "Invalid session backdend: ~:@(~a~)" conf))))))
